--------------------------------------------------------------------------------
--
--  Murphi Model of the Secure Cookie Protocol
--
--------------------------------------------------------------------------------
--
--  version:      1.0
--
--  written by:   Jeremy Brotherton and Kiyoshi Shikuma
--  date:         February 2011
--  affiliation:  Stanford University (Graduate Students)
--
--------------------------------------------------------------------------------
--
--  Protocol is modeled as follows:
-- 
--  1) Client Starts Session (SSL)
--	(Client --{Start}--> Server)
--  2) Server sets a session key (SSL)
--	(Server --{SessionKey} --> Client)
--	3) Client authenticates (SSL)
--	(Client --{Authenticate} -->Server)
--	4) Server sets cookie
--	(Server --{Cookie} --> Client)
--	5) Client uses cookie for server
--	(Client --{Cookie} --> Server)
--
--	The first three steps area always done under SSL,
--  In this model, SSL is just taken to mean that only
--  the sender and intended recipient can read fields
--  in the message other than source and dest
--
--------------------------------------------------------------------------------

--
--
--


--------------------------------------------------------------------------------
-- constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const
SSL:           			false;		-- turn SSL on and off for cookie usage
HighConfidentiality: 	false;		-- turn High confidentiality on and off

NumClients:		 2;   -- number of initiators
NumServers:		 2;   -- number of responders
NumIntruders:    2;   -- number of intruders
NetworkSize:     1;   -- max. number of outstanding messages in network
MaxKnowledge:   20;   -- max. number of messages intruder can remember


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type
	ClientId:		scalarset (NumClients);   -- identifiers
	ServerId:		scalarset (NumServers);
	IntruderId:     scalarset (NumIntruders);
	int_t: 0..99;
  
AgentId:      union {ClientId, ServerId, IntruderId};

MessageType : enum {          -- different types of messages
    M_Start,              			--  {source, dest, user, SSL}
    M_SessionKey,         			--  {source, dest, seskey, SSL}
    M_Authenticate,					--  {source, dest, user, SSL}
    M_SetCookie,					--  all fields set
    M_UseCookie						--  all fields set
};

HMAC : record
	user:			AgentId;
	expire: 		int_t;
	data: 			int_t;		-- Should not be defined if serverkey is defined
	seskey: 		int_t;		-- Should not be defined if serverkey is defined
	serverkey:		int_t;		-- If undefined it is encrypted with another HMAC as key
end;

Cookie : record
	user: 		AgentId;       -- user id of cookie
	expire:     int_t;         -- expiration time of cookie
	data:       int_t;         -- data in cookie
	seskey:     int_t;         -- session key
	serkey:     int_t;         -- server key
	hmac:  		HMAC;          -- hmac in cookie
	hmackey:    HMAC;		   -- hmac used as key k
end;
	
Message : record
	mType:		MessageType;      -- type of message
	source:		AgentId;		  -- source of message
	dest:		AgentId;		  -- intended destination of message
	cookie:		Cookie;			  -- cookie in message
	ssl: 		boolean;		  -- SSL enabled on this message
end;


-- remark: keys, nonces and addresses are encoded in the message by their
--         agent's identifier only. They can be distinguished by their po-
--         sition and the type of the message

ClientStates : enum {
	C_SLEEP,                   -- state after initialization
	C_WAIT_K,                  -- waiting for key from server
	C_WAIT_C,				   -- waiting for cookie from server
	C_COMMIT                   -- client has used cookie
};                             --  (for authentication)

Client : record
	state:     	ClientStates;
	seskey:		int_t;
	cookie: 	Cookie;
	server: 	AgentId;          -- agent who client gets cookie from
	serkey:		int_t;
end;


Server : record
	key:				int_t;						-- Server key never changes
	next_ses_key: 		int_t;
	session_keys: 		array[AgentId] of int_t;
	accepted_conns: 	array[AgentId] of Cookie;				-- Cookie associated with conn
	client:				AgentId;
end;

Intruder : record
	seskey:      array[AgentId] of int_t;              -- known session keys
	cookies:     array[AgentId] of Cookie;             -- known cookies
	servers:   	 array[AgentId] of boolean;            -- servers we sent a M_Start message to
	usernames:   array[AgentId] of AgentId;            -- learned usernames
	expires:     array[AgentId] of AgentId;            -- learned expiration times
	serkey:		 array[AgentId] of int_t;              -- known server private keys
	messages: multiset[MaxKnowledge] of Message;   	   -- known messages
end;
    

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                  -- state variables for
	net: multiset[NetworkSize] of Message;			--  network
	cli: array[ClientId] of Client;     			--  clients
	ser: array[ServerId] of Server;     			--  servers
	int: array[IntruderId] of Intruder;				--  intruders



--------------------------------------------------------------------------------
-- rules
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- behavior of clients

-- client i starts protocol with server or intruder j (step 1)
ruleset i: ClientId do
  ruleset j: AgentId do					
	rule "client starts session (step 1)"

      cli[i].state = C_SLEEP &
      !ismember(j,ClientId) &               -- only servers and intruders
      multisetcount (l:net, true) < NetworkSize

    ==>
    
    var
      outM: Message;   -- new message

    begin
      undefine outM;
      outM.source  := i;  
      outM.dest    := j;  
      undefine outM.cookie;
      outM.mType   := M_Start;			-- First client message
      outM.ssl := true;				-- SSL not yet established

      multisetadd (outM,net);

      cli[i].state     := C_WAIT_K;
      cli[i].server := j;				-- record the client's perception of who it is talking to
    end;
  end;
end;

-- client i client reacts after session key with authentication
ruleset i: ClientId do
  choose j: net do
    rule "client reacts to session key with authentication (step 3)"

      cli[i].state = C_WAIT_K &
      net[j].dest = i &
      (ismember(net[j].source, ServerId) | -- is from server or attacker
      ismember(net[j].source, IntruderId))
    ==>

    var
      outM: Message;   -- outgoing message
      inM:  Message;   -- incoming message

    begin
      inM := net[j];
      multisetremove (j,net);

      if inM.mType=M_SessionKey then   -- correct message type
        undefine outM;
      	outM.source  := i;
      	outM.dest    := inM.source;
      	undefine outM.cookie;
      	outM.mType   := M_Authenticate;
      	outM.cookie.user := i;							-- username sent inside cookie field for authentication
      	outM.cookie.seskey := inM.cookie.seskey;		-- SSL session key sent by server
				cli[i].seskey := inM.cookie.seskey;				-- SSL session key sent by server
				outM.ssl := true;								-- TODO: USE SSL VARIABLE??
	
				multisetadd (outM,net);
	
				cli[i].state     := C_WAIT_C;
      end;
    end;
  end;
end;

-- client i reacts to received cookie
ruleset i: ClientId do
  choose j: net do
    rule "client reacts to received cookie uses it (step 5)"

      cli[i].state = C_WAIT_C &
      net[j].dest = i &
      (ismember(net[j].source,ServerId) | -- is from server or attacker
      ismember(net[j].source, IntruderId))
    ==>

    var
      outM: Message;   -- outgoing message
      inM:  Message;   -- incoming message

    begin
      inM := net[j];
      multisetremove (j,net);

      if inM.mType=M_SetCookie then   -- correct message type
      	cli[i].cookie := inM.cookie;					-- set the cookie on client-side
        undefine outM;
      	outM.source  := i;
      	outM.dest    := inM.source;
      	undefine outM.cookie;
      	outM.cookie := cli[i].cookie;					-- send back the cookie with request
      	outM.mType   := M_UseCookie;
		outM.ssl := SSL;								-- TODO: USE SSL VARIABLE???
--		outM.cookie.expire :=75;
		multisetadd (outM,net);

		cli[i].state     := C_COMMIT;
      end;
    end;
  end;
end;


-- client uses cookie again (step 6+)
ruleset i: ClientId do
  ruleset j: AgentId do					
    rule "client uses cookie again (step 6+)"

      cli[i].state = C_COMMIT &
      !ismember(j,ClientId) &               		-- only servers and intruders
      multisetcount (l:net, true) < NetworkSize

    ==>
    
    var
      outM: Message;   -- new message

    begin
      undefine outM;
      outM.source  := i;  
      outM.dest    := j;  
      undefine outM.cookie;
      outM.cookie := cli[i].cookie;					-- send back the cookie with request
      outM.mType   := M_UseCookie;
			outM.ssl := SSL;								-- TODO: USE SSL VARIABLE???

      multisetadd (outM,net);
    end;
  end;
end;

--------------------------------------------------------------------------------
-- behavior of servers

-- server i reacts to client's session start (step 2)
 ruleset i: ServerId do
  choose j: net do
    rule "server reacts to client's session start (step 2)"

      net[j].dest = i &
      net[j].mType = M_Start &
      (ismember(net[j].source,IntruderId) |
      ismember(net[j].source,ClientId))

    ==>

    var
      outM: Message;   -- outgoing message
      inM:  Message;   -- incoming message

    begin
      inM := net[j];
      multisetremove (j,net);

      undefine outM;
      outM.source  := i;
      outM.dest    := inM.source;			-- identifier of client
      outM.mType   := M_SessionKey;
      undefine outM.cookie;
      outM.cookie.seskey := ser[i].next_ses_key;					-- set new session key for the communitcation
      ser[i].session_keys[outM.dest] := ser[i].next_ses_key;		-- remember the session key for this client
      ser[i].next_ses_key := ser[i].key + 1;						-- generate next session key
      outM.ssl := true;												-- SSL session has been established

      multisetadd (outM,net);

	  	ser[i].client := inM.source;									-- record server's perception of who it is talking to
    end;
  end;
end;

-- Server i reacts to authentication and makes cookie (step 4)
ruleset i: ServerId do
	choose j: net do
		rule "server reacts to authentication request (step 4)"

			net[j].dest = i &
			net[j].mType = M_Authenticate &
			(ismember(net[j].source,IntruderId) |
			ismember(net[j].source, ClientId))

		==>

		var
			outM: Message;   -- outgoing message
			inM:  Message;   -- incoming message

		begin
			inM := net[j];
			multisetremove (j,net);

			undefine outM;
			outM.source  := i;
			outM.dest    := inM.source;								-- identifier of client
			outM.mType   := M_SetCookie;
			undefine outM.cookie;
			outM.cookie.user 	:= inM.cookie.user;					-- username client specified
			--outM.cookie.expire  := inM.cookie.seskey;				-- TODO: HARDCODE EXPIRATION???
			--outM.cookie.data    := inM.cookie.seskey;				-- 
			outM.cookie.expire  := 50;								-- TODO: HARDCODE EXPIRATION???
			outM.cookie.data    := 99;								-- TODO: HARDCODED VALUE???
			outM.cookie.seskey  := inM.cookie.seskey;				-- session key should remain the same
			 
			undefine outM.cookie.hmac;
			undefine outM.cookie.hmackey; 
			         
			outM.cookie.hmac.user  := inM.cookie.user;				-- username client specified
			--outM.cookie.hmac.expire  := inM.cookie.seskey;			-- expiration time
			outM.cookie.hmac.expire  := 50;							-- expiration time
			outM.cookie.hmac.data    := 99;							-- cookie data
			outM.cookie.hmac.seskey  := inM.cookie.seskey;			-- session key
			           
			outM.cookie.hmackey.user := inM.cookie.user;			-- username client specified
			outM.cookie.hmackey.expire  := 50;						-- expiration time
			outM.cookie.hmackey.serverkey  := ser[i].key;			-- static server key
			 
			outM.ssl := true;										-- TODO: USE SSL VARIABLE???
			--ser[i].accepted_conns[inM.cookie.user] := outM.cookie;  -- Keep track of the cookie you sent for HMAC Validation
			multisetadd (outM,net);

		end;
	end;
end;

-- Server i reacts cookie
ruleset i: ServerId do
	choose j: net do
		rule "Server reacts to sent cookie (step 6)"

				net[j].dest = i &
				net[j].mType = M_UseCookie &
				(ismember(net[j].source,IntruderId) |
				ismember(net[j].source, ClientId))

			==>

			var
				outM: Message;   -- outgoing message
				inM:  Message;   -- incoming message

			begin
				alias inM: net[j] do   -- incoming message
				--ser[i].accepted_conns[inM.source] := inM.cookie;
				
				-- MOVED THIS TO AN INVARIANT
				if (inM.cookie.hmackey.serverkey=ser[i].key & -- Check hmackey validity
						inM.cookie.hmackey.user = inM.cookie.user &
						inM.cookie.hmackey.expire = inM.cookie.expire) & -- hmackey valid so 
						( inM.cookie.hmac.user = inM.cookie.user & -- check hmac
						inM.cookie.hmac.expire = inM.cookie.expire &
						inM.cookie.hmac.data = inM.cookie.data &
						inM.cookie.hmac.seskey = ser[i].session_keys[inM.source])
				then   -- cookie is okay, accept connection
					ser[i].accepted_conns[inM.source] := inM.cookie;
				else
					error "cookie integrity or anti-replay holds"
				end;
				multisetremove (j,net);
			end;
		end;
	end;
end;

--------------------------------------------------------------------------------
-- behavior of intruders
-- intruder i intercepts messages
 ruleset i: IntruderId do
  choose j: net do
    rule "intruder intercepts"

      !ismember (net[j].source, IntruderId)    -- not for intruders' messages
      ==>

      var
        temp: Message;
				test: boolean;
    begin
      alias msg: net[j] do   -- message to intercept
				
				if !msg.ssl | msg.dest = i then   -- Not SSL protected message or
					--directed to us so cna read internal
					alias messages: int[i].messages do
						temp := msg;
						if msg.mType = M_SetCookie | msg.mType = M_UseCookie then
							int[i].cookies[msg.cookie.user] := temp.cookie;
						end;
						--can blindly copy message
						undefine temp.source;   -- delete useless information
						undefine temp.dest;
						if ( multisetcount (l:messages, -- add only if not there
										messages[l].mType = temp.mType &
										messages[l].ssl = temp.ssl &
										messages[l].source = msg.source &
										messages[l].dest = msg.dest
									) = 0 & 
									multisetcount (l:int[i].messages, true) < MaxKnowledge ) then
							multisetadd (temp, int[i].messages);
						end;
					end;
				end;
				
					
				multisetremove (j,net);
			end;
		end;
	end;
end;

-- intruder i sends recorded message
ruleset i: IntruderId do         -- arbitrary choice of
  choose j: int[i].messages do   --  recorded message
    ruleset k: AgentId do        --  destination
      rule "intruder sends recorded message"

        !ismember(k, IntruderId) &                 -- not to intruders
        multisetcount (l:net, true) < NetworkSize

      ==>

      var
        outM: Message;

      begin
        outM        := int[i].messages[j];
        outM.source := i;
        outM.dest   := k;
		if outM.mType = M_Start & ismember(k, ServerId) then
			int[i].servers[k] := true;
		end;
		--if outM.mType = M_SetCookie then
		--	outM.cookie.user := i;
		--end;
			

        multisetadd (outM,net);
      end;
    end;
  end;
end;

-- intruder i generates message with its own username
--ruleset i: IntruderId do       -- arbitrary choice of
--	ruleset j: AgentId do         --  destination = key
--		ruleset l: MessageType do    --  message type
--			ruleset m: AgentId do       --  usernames
--						rule "intruder generates message"
--
--						!ismember(j, IntruderId) &       -- not to intruders
--						!isundefined(int[i].usernames[m]) &     -- usernames must be known
--						multisetcount (t:net, true) < NetworkSize
--
--						==>
--						
--						var
--						outM: Message;
--						
--						begin
--							undefine outM;
--							outM.source := i;
--							outM.dest   := j;
--							outM.mType  := l;
--							
--							switch l   -- set content dependent on message type
--								case M_Start:
--									outM.mType   := M_Start;			-- First client message
--									outM.ssl := SSL;				-- SSL not yet established
--						end;
--						
--						--multisetadd (outM,net);
--						end;
--					end;
--		end;
--	end;
--end;

--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate
  -- initialize clients
  undefine cli;
  for i: ClientId do
    cli[i].state     := C_SLEEP;
    cli[i].server := i;
    undefine cli[i].seskey;
    undefine cli[i].cookie;
  end;

  -- initialize servers
  undefine ser;
  for i: ServerId do
    ser[i].key := 1;
    ser[i].next_ses_key := 1;
    for j: AgentId do
    	ser[i].session_keys[j] := 0;
    	undefine ser[i].accepted_conns[j];
    	undefine ser[i].accepted_conns[j].user;
    end
  end;

  -- initialize intruders
  undefine int;
	for i: IntruderId do   -- the only nonce known is the own one
		for j: AgentId do  
			int[i].servers[j] := false;
		end;
		int[i].usernames[i] := i;	-- You know your own username
	end;

  -- initialize network 
  undefine net;
end;



--------------------------------------------------------------------------------
-- invariants
--------------------------------------------------------------------------------

-- Authentication Invariants
--invariant "Client authenticated by server correctly"
--	forall i: ServerId do
--		forall j: AgentId do
--			!ISUNDEFINED(ser[i].accepted_conns[j].user)
--			->
--			ser[i].accepted_conns[j].user = j
--		end
--	end;


--invariant "Server connections intended by clients" -- check that the client server thought it was talking to also thinks it was talking to this server
--	forall i: ServerId do
--		forall j: ClientId do
--			!ISUNDEFINED(ser[i].accepted_conns[j].user)
--			->
--			cli[j].server = i &
--			( cli[j].state = C_WAIT_C |
--		  cli[j].state = C_COMMIT)
--		end
--	end;
	
-- Confidentiality Invariants
--invariant "low confidentiality holds"
--	forall i: IntruderId do
--		forall j: ClientId do
--			!HighConfidentiality & cli[j].state = C_COMMIT 
--			-> isundefined(int[i].cookies[j].data)
--	end
--end;

--invariant "high confidentiality holds"
--	forall i: IntruderId do
--		forall j: ClientId do
--			forall k: ServerId do
--				HighConfidentiality & cli[j].state = C_COMMIT 
--				-> isundefined(int[i].serkey[k]) & isundefined(cli[j].serkey)
--			end
--		end
--	end;
		
-- Integrity & anti-replay invariants at line 413, 414